"""
This module is under development and still not in
use.
"""
from typing import Union


class FrameAdapter:
    """
    The adapter.
    """

    def __init__(
        self,
        frame: 'Frame',
        # # TODO: Maybe 'RenderContext' instead (?)
        # opengl_context: 'moderngl.Context'
    ):
        self.frame: 'Frame' = frame
        """
        The frame to be adapted.
        """
        # self.opengl_context: 'moderngl.Context' = opengl_context
        # """
        # The context itself.
        # """

    """
    Operations here below
    """
    def resize_with_padding(
        self,
        size: tuple[int, int]
    ) -> 'FrameAdapter':
        self.frame.pending_operations.append(ResizeWithPaddingOperation(size))

        return self

    def to_gpu(
        self,
        render_context: 'RenderContext'
    ):
        return self._resolve(
            backend = 'gpu',
            render_context = render_context
        )

    def to_cpu(
        self
    ):
        return self._resolve(
            backend = 'cpu'
        )

    def _resolve(
        self,
        # TODO: This 'backend' should be more strict
        backend: str,
        render_context: Union['RenderContext', None]
    ):
        if backend not in ['cpu', 'gpu']:
            raise Exception('The "backend" provided is not valid.')
        
        resource = self.frame.resource

        from yta_editor_frame.frame.resource.identifier import FrameResourceIdentifier

        if (
            backend == 'gpu' and
            (
                FrameResourceIdentifier.is_gpu_render_target(self.frame.resource) or
                FrameResourceIdentifier.is_gpu(self.frame.resource)
            )
        ):
            texture = (
                self.frame.resource.frame.texture
                if FrameResourceIdentifier.is_gpu_render_target(self.frame.resource) else
                self.frame.resource.frame
            )
        elif backend == 'gpu':
            # 1. Get texture from pool
            texture = render_context._render_resources.texture.acquire(
                width = self.frame.width,
                height = self.frame.height,
                number_of_components = 4,
                dtype = 'f1'
            )
            # 2. Get the numpy array from frame
            numpy_array = (
                self.frame.resource.frame
                if FrameResourceIdentifier.is_cpu(self.frame.resource) else
                self.frame.resource.frame.to_ndarray(format = 'rgba')
            )
            # 3. Write numpy array into texture
            texture.write(numpy_array)
        elif (
            backend == 'cpu' and
            (
                FrameResourceIdentifier.is_cpu_av_videoframe(self.frame.resource) or
                FrameResourceIdentifier.is_cpu(self.frame.resource)
            )
        ):
            numpy_array = (
                self.frame.resource.frame
                if FrameResourceIdentifier.is_cpu(self.frame.resource) else
                self.frame.resource.frame.to_ndarray(format = 'rgba')
            )
        else:
            # 1. Get texture
            texture = (
                self.frame.resource.frame
                if FrameResourceIdentifier.is_gpu(self.frame.resource) else
                self.frame.resource.frame.texture
            )
            # 2. Read as numpy array
            numpy_array = texture.read()

        # aplicar pipeline
        for pending_operation in self.frame.pending_operations:
            if backend == 'gpu':
                texture = pending_operation.apply_gpu(
                    frame = texture,
                    render_context = render_context
                )
            else:
                numpy_array = pending_operation.apply_cpu(
                    frame = numpy_array
                )

        return (
            texture
            if backend == 'gpu' else
            numpy_array
        )

class FrameToBeAdapted:
    """
    Class to wrap the functionality that adapts the
    frames.
    """

    def __init__(
        self,
        frame: 'Frame'
    ):
        self.frame: 'Frame' = frame
        """
        The frame to be adapted.
        """
        self.operations_pending = []
        """
        The operations that are pending to be done.
        """


from yta_editor_frame.frame import Frame

import numpy as np


class _FrameOperation:
    """
    *For internal use only*

    Class to indicate an operation that must be done
    within a frame.
    """

    def apply_gpu(
        self,
        frame: 'moderngl.Texture',
        render_context: 'RenderContext'
    ) -> 'moderngl.Texture':
        """
        Perform this frame operation by using the GPU.
        """
        raise NotImplementedError

    def apply_cpu(
        self,
        frame: 'np.ndarray'
    ) -> 'np.ndarray':
        """
        Perform this frame operation by using the CPU.
        """
        raise NotImplementedError
    
class ResizeWithPaddingOperation(_FrameOperation):
    """
    Operation to resize a frame with padding.
    """

    def __init__(
        self,
        size: tuple[int, int]
    ):
        self.size: tuple[int, int] = size
        """
        The expected size of the frame after adapting it.
        """

    def apply_gpu(
        self,
        texture: 'moderngl.Texture',
        render_context: 'RenderContext'
    ):
        width_output, height_output = self.size
        width_input, height_input = texture.size

        scale_x = width_input / width_output
        scale_y = height_input / height_output

        opengl_context = render_context._render_resources.opengl_context

        # TODO: Maybe a 'RenderTarget' is better (?)
        # render_target = render_context._render_resources._render_target_provider.acquire(
        #     width = width_output,
        #     height = height_output,
        #     number_of_components = 4,
        #     dtype = 'f1'
        # )
        output_texture = render_context._render_resources.texture.acquire(
            width = width_output,
            height = height_output,
            number_of_components = 4,
            dtype = 'f1'
        )
        # output_tex = ctx.texture((width_output, height_output), 4)
        fbo = opengl_context.framebuffer([output_texture])

        fbo.use()
        opengl_context.clear(0, 0, 0, 0)

        shader = get_cached_shader(ctx, "resize_pad")

        shader["scale"].value = (scale_x, scale_y)

        texture.use(0)
        render_fullscreen_quad(shader)

        return output_texture
    
    def apply_cpu(
        self,
        frame: 'Frame'
    ):
        numpy_array = _adapt_with_alpha(
            frame = frame.as_numpy_ndarray(),
            size = self.size
        )

        frame = Frame.from_frame(numpy_array)
        """
        We added and alpha background, so we have transparent
        pixels and the alpha layer.
        """
        frame.metadata.has_transparent_pixels = True
        frame.metadata.is_fully_transparent = False
        frame.metadata.is_fully_opaque = False

        return frame

        return super().apply_cpu(frame)
    
"""
Utils below
"""
def _adapt_with_alpha(
    frame: np.ndarray,
    size: tuple[int, int]
) -> np.ndarray:
    """
    Creates a fully alpha background with the `size`
    given and puts the `frame` provided in the middle of
    it.

    This is the functionality to be done when using the
    `ALPHA_BACKGROUND` option.
    """
    if frame.ndim != 3:
        raise Exception('The frame must be HxWxC.')

    src_h, src_w, src_c = frame.shape

    if src_c not in (3, 4):
        raise Exception('The frame must be RGB or GBA.')

    # Final transparent frame
    dst = np.zeros((size[1], size[0], 4), dtype = frame.dtype)

    # RGBA if necessary
    if src_c == 3:
        src_rgba = np.empty((src_h, src_w, 4), dtype = frame.dtype)
        src_rgba[..., :3] = frame
        src_rgba[..., 3] = np.iinfo(frame.dtype).max
    else:
        src_rgba = frame

    # Offsets to center
    offset_x = (size[0] - src_w) // 2
    offset_y = (size[1] - src_h) // 2

    if (
        offset_x < 0 or
        offset_y < 0
    ):
        # TODO: Do rescale and crop if needed
        raise Exception('Original frame is bigger than the size requested.')

    # Copy it
    dst[
        offset_y:offset_y + src_h,
        offset_x:offset_x + src_w
    ] = src_rgba

    # Force it to be continuous to avoid PyAv errors
    dst = np.ascontiguousarray(dst)

    return dst

# TODO: Create the GPU node that does this



from yta_editor.tracks.items.transition.transition_node_registry import TransitionNodeRegistry
from yta_editor.tracks.items.transition.transition_specification import TransitionSpecification
from yta_editor_frame.frame import Frame
from yta_editor.media.video import VideoFileMedia
from yta_editor_frame.helper import FrameHelper
from yta_editor_utils.texture import TextureUtils
from yta_validation import PythonValidator
from yta_math_easings.abstract import EasingFunction
from typing import Union

import numpy as np


# TODO: This class below could be repeated and I just
# created because I needed to (at least by now)
# TODO: The progress is not being calculated correctly
class _TransitionCalculator:
    """
    *For internal use only*

    A class to simplify the way we make the calculations
    related to a transition. This class must be used within
    a transition item that needs these kind of calculations.
    """

    @property
    def fps(
        self
    ) -> float:
        """
        The `fps` associated to the `transition_item` that
        is attached to this instance.
        """
        # TODO: Is this the 'output_fps' (?)
        return self._transition_item.fps
    
    @property
    def output_size(
        self
    ) -> float:
        """
        The `output_size` associated to the `transition_item`
        that is attached to this instance.
        """
        return self._transition_item.output_size

    def __init__(
        self,
        render_context: 'RenderContext',
        transition_item: 'TransitionTrackItem',
        duration: float,
        transition_specification: Union['TransitionSpecification', None] = None
    ):
        self.render_context: 'RenderContext' = render_context
        """
        The `RenderContext` instance.
        """
        self._transition_item: 'TransitionTrackItem' = transition_item
        """
        The transition item this helper was instantiated for.
        """
        self.duration: float = duration
        """
        The duration of the transition.
        """

        transition_specification = (
            TransitionSpecification(
                node_name = TransitionNodeRegistry.get_default_key(),
                # TODO: This must be dynamic according to the default
                parameters = {
                    'border_smoothness': 0.02
                },
                easing_function = 'linear'
            )
            if transition_specification is None else
            transition_specification
        )

        if not TransitionNodeRegistry.is_registered(transition_specification.node_name):
            raise Exception(f'There is no transition node processor registered with the given "{transition_specification.node_name}" name.')
        
        # TODO: The parameters of the specification must be
        # validated here according to the node processor
        # specification (which doesn't exist yet). By now
        # we are not doing this parameters validation
        self._transition_specification: 'TransitionSpecification' = transition_specification
        """
        The specification that defines the node processor
        and the parameters we should use.
        """

        # The transition could need the 'mask_input'
        mask_input_filename = self._transition_specification.parameters.get('mask_input', None)
        self._mask_video: Union[VideoFileMedia, None] = (
            VideoFileMedia(
                filename = mask_input_filename
            )
            if mask_input_filename is not None else
            None
        )
        """
        *For internal use only*

        The alpha video that we will use to read the
        frames and apply the mask.
        """
        # TODO: What about the GPU or CPU (?)
        self._node_processor = TransitionNodeRegistry.get(self._transition_specification.node_name)
        """
        *For internal use only*

        The node processor that we will use to process
        the input frame with the corresponding mask
        frames.
        """

    # TODO: This method can be overwritten to change
    # its behaviour if the specific transition needs
    # it
    def process_frame(
        self,
        frame_a: Frame,
        frame_b: Frame,
        evaluation_context: 'EvaluationContext'
    ) -> Frame:
        """
        Process the `frame_a` and the `frame_b` provided and
        apply the transition at the `t_progress` provided by
        using the node of the `TransitionTrackItem` associated
        to this instance.
        """
        # TODO: Maybe this can be placed in the general
        # class if it doesn't change

        # TODO: This can be a texture
        processed_frame = self._get_process_frame_array(
            frame_a = frame_a,
            frame_b = frame_b,
            evaluation_context = evaluation_context
        # TODO: Force this with the 'yta-numpy' utils to 'np.uint8' (?)
        )

        # frame_a.release()
        # frame_b.release()

        # I think it is uint8 always...
        return processed_frame
    
    # TODO: Overwrite this method to make it custom (?)
    def _get_process_frame_array(
        self,
        frame_a: Frame,
        frame_b: Frame,
        evaluation_context: 'EvaluationContext'
    ) -> 'Frame':
        """
        Get the numpy array that will be used to build the
        pyav VideoFrame that will be returned.
        """
        inputs = {
            'first_input': frame_a,
            'second_input': frame_b
        }

        # TODO: Check this below to improve it if possible
        parameters_dynamic = self._transition_specification.parameters.copy()

        # Alpha mask transition, we need to handle it in
        # a special way
        if self._mask_video is not None:
            # TODO: Make it a property (?)
            easing_function = EasingFunction.get(self._transition_specification.easing_function)()
            # We calculate the transition t_media with the progress
            t_transition_media = easing_function.ease(evaluation_context.progress.value) * (self._mask_video.number_of_frames - 1) / self._mask_video.fps

            # We get a VideoFrame, so we transform it
            """
            TODO: This below is working but maybe it is not
            the best way. An alpha video should be short and
            maybe we can load it in memory directly or 
            something. Try to improve it, but remember: it
            is working :)
            """
            # TODO: Maybe transform it in other way for optimization
            inputs['mask_input'] = Frame.from_frame(
                frame = self._mask_video.get_video_frame_at(
                    t_media = t_transition_media
                ).to_ndarray(format = 'rgba'),
                render_context = self.render_context
            )

            # We don't want it as a dynamic parameter but as input
            del parameters_dynamic['mask_input']

        render_target = self.render_context.gpu.render_target.acquire(
            width = self.render_context.config.size[0],
            height = self.render_context.config.size[1],
            number_of_components = 4,
            dtype = 'f1'
        )

        frame_processed = self._node_processor.process(
            render_target = render_target,
            inputs = inputs,
            evaluation_context = evaluation_context,
            # TODO: This is how we were using the easing function before
            # progress = self.easing_function.ease(float(evaluation_context.progress)),
            # TODO: Remove when everything is working
            # output_size = self.output_size,
            # Parameters will be validated in
            parameters = parameters_dynamic
        )

        return frame_processed
    
    def _get_process_audio_frames(
        self,
        audio_frames_a: list['AudioFrameWrapped'],
        audio_frames_b: list['AudioFrameWrapped'],
        evaluation_context: 'EvaluationContext'
    ) -> list['AudioFrame']:
        """
        Get the numpy array that will be used to build the
        pyav AudioFrame that will be returned.
        """

        """
        TODO: I don't know how to handle the audio within the
        transitions, because each transition works in a specific
        way and making the audio behave like the video movement
        is not easy... It could be based on the 't_progress',
        using something like this below (based on S-Curve Crossfade
        weights):

        weight_a = 0.5 * (1 + math.cos(t_progress * math.pi))
        weight_b = 1 - weight_a
        mixed_audio_frames = audio_frames_a * weight_a + audio_frames_b * weight_b

        But, as I don't know actually how to do it, by now we are
        just mixing both of them during the transition time in the
        same way as the video frames are accessed.
        """
        # TODO: Move this to top (?)
        from yta_editor.utils.frame_combinator import AudioFrameCombinator
        from yta_editor.timeline import concatenate_audio_frames

        collapsed_frames = [
            concatenate_audio_frames(frames)
            for frames in [audio_frames_a, audio_frames_b]
        ]

        # We keep only the non-silent frames because
        # we will sum them after and keeping them will
        # change the results.
        non_empty_collapsed_frames = [
            frame._frame
            for frame in collapsed_frames
            if not frame.is_from_gap
        ]

        if len(non_empty_collapsed_frames) == 0:
            # If they were all silent, just keep one
            non_empty_collapsed_frames = [collapsed_frames[0]._frame]

        frames = [
            AudioFrameCombinator.sum_tracks_frames(
                tracks_frames = non_empty_collapsed_frames,
                sample_rate = self._transition_item._track.audio_fps,
                # TODO: This was not being sent before
                layout = self._transition_item._track.audio_layout,
                format = self._transition_item._track.audio_format
            )
        ]

        # TODO: Do we need the '**self._transition_specification.parameters'
        # for the audio frames (?)

        # for audio_frame in frames:
        #     yield audio_frame
        # TODO: Should I yield (?)
        return frames

# TODO: No longer used, remove it (?)
def frame_to_uint8_ndarray(
    frame: Union['moderngl.Texture', 'np.ndarray'],
    do_include_alpha: bool = True
):
    """
    Force the `frame` provided to be transformed into
    a uint8 numpy ndarray, removing the alpha channel
    if existing and `do_include_alpha` is `False`.
    """
    if (
        not PythonValidator.is_instance_of(frame, 'Texture') and
        not PythonValidator.is_numpy_array(frame)
    ):
        # TODO: Wtf? This should not happen
        raise Exception('Wowowowooo, no texture, no np.ndarray... wtf?')
    
    if PythonValidator.is_instance_of(frame, 'Texture'):
        # Old way below
        # frame, has_alpha = FrameHelper.texture_to_ndarray(frame)

        # if (
        #     has_alpha and
        #     not do_include_alpha
        # ):
        #     frame = frame[:, :, :3]

        frame = FrameHelper.texture_to_ndarray(frame)

        # if (
        #     not do_include_alpha and
        #     frame.ndim == 3 and
        #     frame.shape[2] > 3
        # ):
        #     frame = frame[:, :, :3]

    return TextureUtils.numpy_to_uint8(frame)


from yta_editor.media.video import VideoFileMedia
from yta_editor.transformations.effects.abstract import Effect
from yta_editor_frame.frame import Frame
from yta_editor_time.evaluation_context import EvaluationContext
from yta_editor_frame.frame.resource.identifier import FrameResourceIdentifier
from yta_editor_frame.decorators import force_frame
from yta_editor.utils import try_release_frame
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from typing import Union
from abc import abstractmethod


class VideoEffect(Effect):
    """
    An effect for a video input.

    This class is just to be able to identify the
    hierarchy and which effect is designed for 
    the video.
    """

    _parameter_values_to_skip_process: dict[str, any] = {}
    """
    The dict that contains the parameter keys and
    the values that must have to skip the
    processing if all the parameters included here
    have the expected values. A single value that
    has a different value or if the dict is empty,
    the processing must be done and not skipped.
    """

    @property
    @abstractmethod
    def do_need_render_target(
        self
    ) -> bool:
        """
        Boolean flag to indicate if the effect needs a
        `RenderTarget` instance to write on it because it
        will use at least one node that works in GPU.
        """
        pass

    def __init__(
        self,
        # TODO: Set the type (VideoEditorParameter)
        render_context: 'RenderContext',
        parameters: dict[str, any]
    ):
        self._parameters: dict[str, any] = parameters
        """
        The parameters that must be evaluated to be able to
        apply the effect in the video, and the keys must be
        exactly the values the nodes are expecting.
        """
        self.render_context: 'Rendercontext' = render_context
        """
        The `RenderContext` instance.
        """

    def _get_params_at(
        self,
        evaluation_context: EvaluationContext
    ) -> dict:
        """
        *For internal use only*

        Get the parameters that must be applied at the given
        `evaluation_context`.
        """
        ParameterValidator.validate_mandatory_instance_of('evaluation_context', evaluation_context, EvaluationContext)

        parameters = {}
        for parameter_name, parameter_value in self._parameters.items():
            parameters[parameter_name] = parameter_value.evaluate(
                evaluation_context = evaluation_context
            )

        return parameters
    
    def _do_skip_process(
        self,
        parameters: dict[str, any]
    ) -> bool:
        """
        *For internal use only*

        Check if the `parameters` provided have the
        values that make the frame change not, so we
        can avoid applying the effect and returning it
        as it is directly. If a single parameter has a
        different value that the one to skip
        processing the frame, we must process it and
        this method will return `False`.

        This method must be set for every specific
        effect to avoid calculations if we don't need
        them.
        
        Here you have on example:

        ```
        parameters = {
            'factor': 1.0
        }
        ```
        """
        if len(self._parameter_values_to_skip_process.keys()) == 0:
            return False
        
        for parameter_key, parameter_value in self._parameter_values_to_skip_process.items():
            if parameters[parameter_key] != parameter_value:
                return False

        return True

    @abstractmethod
    def _apply_effect(
        self,
        frame: 'Frame',
        render_target: Union['RenderTarget', None],
        parameters: dict[str, any],
        evaluation_context: EvaluationContext,
    ) -> Frame:
        """
        *For internal use only*

        Execute the node processors we need to generate
        the desired effect, using the `inputs` and the
        `parameters` provided to obtain a processed and
        final frame.

        This method must be overwritten in the specific
        effects to do the changes we need.
        """
        pass

    @force_frame('frame')
    def apply(
        self,
        frame: 'Frame',
        render_target: Union['RenderTarget', None],
        evaluation_context: EvaluationContext,
    ) -> 'Frame':
        """
        Apply the effect to the given `frame` according to
        the `evaluation_context`.
        """
        # 1. Evaluate the parameters
        parameters = self._get_params_at(
            evaluation_context = evaluation_context
        )

        # 2. Skip if parameters doesn't make any change
        if self._do_skip_process(parameters):
            return frame
        
        # 3. Process the effect
        frame_modified = self._apply_effect(
            frame = frame,
            render_target = render_target,
            parameters = parameters,
            evaluation_context = evaluation_context
        )

        return frame_modified
    
    def copy(
        self
    ) -> 'VideoEffect':
        """
        Get a copy of this `VideoEffect` instance.
        """
        return VideoEffect(
            render_context = self.render_context,
            parameters = self._parameters.copy()
        )
    
class VideoEffectSingleNode(VideoEffect):
    """
    A VideoEffect that is applied by using a
    single node processor for a single frame.
    """

    @property
    def do_need_render_target(
        self
    ):
        return self._node_processor._do_use_gpu

    def __init__(
        self,
        # TODO: Set type
        node_processor: any,
        parameters: dict[str, any],
        render_context: 'RenderContext'
    ):
        # TODO: Validate that is a NodeProcessor (?)
        self._node_processor: any = node_processor(
            render_context = render_context
        )
        """
        *For internal use only*

        The node processor that will modify the frame.
        """

        # TODO: Validate that are ok for the specific node
        super().__init__(
            parameters = parameters,
            render_context = render_context
        )

class VideoEffectSingleNodeBaseInput(VideoEffectSingleNode):
    """
    A VideoEffect that is applied by using a
    single node processor for a single frame.

    This effect will use a single input `base_input`
    with the frame provided, and apply the changes
    based on the `parameters` given.
    """

    def _apply_effect(
        self,
        frame: 'Frame',
        render_target: Union['RenderTarget', None],
        parameters: dict[str, any],
        evaluation_context: EvaluationContext
    ) -> Frame:
        """
        *For internal use only*

        Process the `frame` provided with the `parameters`
        given, according to the specific `node` that this
        specific class is using.

        This method will return a `np.ndarray` or a
        `moderngl.Texture` (according to if it is processed
        with CPU or GPU) that must be transformed into a
        `Frame` instance.
        """
        return self._node_processor.process(
            render_target = render_target,
            inputs = {
                'base_input': frame
            },
            evaluation_context = evaluation_context,
            parameters = parameters
        )
    
# TODO: Maybe we could implement this directly in
# the 'alphapedia_mask_reveal.py' because we don't
# have any other node
class VideoEffectSingleNodeBaseInputAndMaskInput(VideoEffectSingleNode):
    """
    A VideoEffect that is applied by using a
    single node processor for a single frame
    combined with a mask frame.

    This effect will use a single input `base_input`
    with the frame provided and a 'mask_input' to
    combine with, and apply the changes based on the
    `parameters` given.
    """

    def __init__(
        self,
        # TODO: Set type
        node_processor: any,
        mask_input_filename: str,
        parameters: dict[str, any],
        render_context: 'RenderContext'
    ):
        self._mask_video: VideoFileMedia = VideoFileMedia(
            filename = mask_input_filename
        )
        """
        *For internal use only*

        The alpha video that we will use to read the
        frames and apply the mask.
        """

        super().__init__(
            node_processor = node_processor,
            parameters = parameters,
            render_context = render_context
        )

    """
    TODO: We are mixing parameters and inputs, so maybe
    I could refactor this a bit to respect more the
    input and parameter concepts.
    """
    def _get_params_at(
        self,
        evaluation_context: EvaluationContext
    ):
        # Get parameters as usual
        parameters = super()._get_params_at(
            evaluation_context = evaluation_context
        )

        # TODO: This is repeated in the 'transition_calculator.py'
        # TODO: What about an easing function (?)
        t_transition_media = evaluation_context.progress.value * (self._mask_video.number_of_frames - 1) / self._mask_video.fps

        video_frame = self._mask_video.get_video_frame_at(
            t_media = t_transition_media
        ).to_ndarray(format = 'rgba')
        
        video_frame = (
            video_frame.frame
            if PythonValidator.is_instance_of(video_frame, 'VideoFrame') else
            video_frame
        )

        # Get extra input
        parameters['mask_frame'] = video_frame

        return parameters
        
    def _apply_effect(
        self,
        frame: 'Frame',
        render_target: Union['RenderTaregt', None],
        parameters: dict[str, any],
        evaluation_context: EvaluationContext
    ) -> Frame:
        """
        *For internal use only*

        Process the `frame` provided with the `parameters`
        given, according to the specific `node` that this
        specific class is using.

        This method will return a `np.ndarray` or a
        `moderngl.Texture` (according to if it is processed
        with CPU or GPU) that must be transformed into a
        `Frame` instance.
        """
        return self._node_processor.process(
            render_target = render_target,
            inputs = {
                'base_input': frame,
                # We don't want the input as a parameter => '.pop'
                'mask_input': parameters.pop('mask_frame')
            },
            evaluation_context = evaluation_context,
            parameters = parameters
        )

class VideoEffects:
    """
    A set of effects to apply in a video input.
    """

    # TODO: This is not as simple as a list, it must be
    # managed with the specific nodes we designed that
    # are able to combine and are time-concerned.

    @property
    def render_context(
        self
    ) -> 'RenderContext':
        """
        The `RenderContext` instance associated to this,
        extracted from the first effect available.
        """
        return self.effects[0].render_context
    
    def __init__(
        self,
        effects: list[VideoEffect] = []
    ):
        self.effects: list[VideoEffect] = effects
        """
        The list of the effects to apply in the video.
        """

    def __getitem__(
        self,
        index: int
    ):
        return self.effects[index]

    def __len__(
        self
    ):
        return len(self.effects)

    def __iter__(
        self
    ):
        return iter(self.effects)

    @force_frame('frame')
    def apply(
        self,
        frame: 'Frame',
        evaluation_context: EvaluationContext,
        output_size: tuple[int, int]
    ) -> 'Frame':
        """
        Apply all the effects that should be applied at the
        given `t` time moment to the provided audio `frame`.
        """
        if len(self) > 0:
            """
            I need to handle the RenderTarget only if needed,
            because if it is being processed with CPU I don't need
            any RenderTarget, but if processed with GPU I'd
            probably need a ping-pong.

            The problem is that an Effect could be consecutive
            effects (nodes) that handle with GPU/CPU mixed...
            
            Only consecutive GPU effects must handle different RTs.

            If effects are parallel I need more than ping pong, but
            by now this is not happening.
            """
            if len(self.effects) == 1:
                effect = self.effects[0]
                
                if effect.do_need_render_target:
                    # GPU
                    render_target_effect = self.render_context.gpu.render_target.acquire(
                        width = output_size[0],
                        height = output_size[1],
                        number_of_components = self.render_context.config.number_of_components,
                        dtype = self.render_context.config.dtype
                    )

                    """
                    This 'frame_modified' could be exactly the
                    'frame' provided if the effect does not have
                    to be applied actually.
                    """
                    # TODO: Need to detect if effect is with CPU or not
                    frame_modified = self.effects[0].apply(
                        frame = frame,
                        render_target = render_target_effect,
                        evaluation_context = evaluation_context
                    )

                    if frame_modified is not frame:
                        try_release_frame(
                            frame = frame,
                            extra_message = f'in "VideoEffects.apply() with evaluation_context.t={str(float(evaluation_context.t))}'
                        )
                        frame = frame_modified
                else:
                    # CPU
                    frame_modified = self.effects[0].apply(
                        frame = frame,
                        render_target = None,
                        evaluation_context = evaluation_context
                    )

                    frame = frame_modified
            else:
                """
                If at least one effect needs GPU, we need to
                handle the `RenderTarget`
                """
                number_of_gpu_effects: int = sum(
                    effect.do_need_render_target
                    for effect in self.effects
                )

                """
                TODO: This could be simplified and if it is 1
                effect, acquire only one render target, but by
                now I act as I was doing before (that was 
                working correctly).
                """
                if number_of_gpu_effects > 0:
                    # GPU, at least one effect is based on it
                    # len(self.effects) will be always >1 here
                    render_target_1, render_target_2 = self.render_context.gpu.render_target.acquire_ping_pong(
                        width = output_size[0],
                        height = output_size[1],
                        number_of_components = self.render_context.config.number_of_components,
                        dtype = self.render_context.config.dtype
                    )

                    render_target_to_write = render_target_1
                    current = frame

                    for video_effect in self.effects:
                        if video_effect.do_need_render_target:
                            output = video_effect.apply(
                                frame = current,
                                render_target = render_target_to_write,
                                evaluation_context = evaluation_context
                            )

                            if output is current:
                                # Passthrough → no swap, no changes
                                continue

                            current = output

                            # Ping-pong swap
                            render_target_to_write = (
                                render_target_2
                                if render_target_to_write is render_target_1
                                else render_target_1
                            )
                        else:
                            output = video_effect.apply(
                                frame = current,
                                render_target = None,
                                evaluation_context = evaluation_context
                            )

                            if output is current:
                                # Passthrough → no swap, no changes
                                continue

                            current = output

                    frame = current

                    # Lets free resources
                    final_render_target = None
                    if (
                        hasattr(frame, 'resource') and
                        hasattr(frame.resource, 'frame') and
                        FrameResourceIdentifier.is_gpu_render_target(frame.resource)
                    ):
                        if frame.resource.frame in (render_target_1, render_target_2):
                            final_render_target = frame.resource.frame

                    # Free the 'RenderTarget' that is not in the final frame
                    if final_render_target is render_target_1:
                        self.render_context.gpu.render_target.release(render_target_2)
                    elif final_render_target is render_target_2:
                        self.render_context.gpu.render_target.release(render_target_1)
                    else:
                        # No effect applied, no RT used...
                        self.render_context.gpu.render_target.release(render_target_1)
                        self.render_context.gpu.render_target.release(render_target_2)
                else:
                    # CPU (all nodes are CPU-based)
                    current = frame

                    for video_effect in self.effects:
                        output = video_effect.apply(
                            frame = current,
                            render_target = None,
                            evaluation_context = evaluation_context
                        )

                        if output is current:
                            # Passthrough → no swap, no changes
                            continue

                        current = output

                    frame = current

        return frame

    def copy(
        self
    ) -> 'VideoEffects':
        """
        Get a copy of this `VideoEffects` instance.
        """
        return VideoEffects(
            effects = [
                video_effect.copy()
                for video_effect in self.effects
            ]
        )

# TODO: This should go to a utils I think
def _should_use_gpu(
    # TODO: Set type
    node_processor: any,
    do_use_gpu: bool = True
) -> bool:
    """
    *For internal use only*

    Check if we should use GPU or not, according
    to what the user is requesting with the
    `do_use_gpu` parameter but checking if it is
    available in the `node_processor` or not.

    This method is useful to know if we have to
    transform a frame to a specific format (for
    GPU or CPU) before sending it as an argument.
    """
    # TODO: Validate 'node_processor'
    return (
        (
            do_use_gpu and
            node_processor.is_gpu_available
        )
    ) or (
        (
            not do_use_gpu and
            not node_processor.is_cpu_available
        )
    )


"""
Our awesome editor module in which we
have all the classes that interact with
it and make it possible.

An editor includes a single timeline,
built by tracks, in which we place media
elements and we are able to apply effects
to them.

Here is a brief explanation about the
hierarchy:

- The editor has a timeline.
- The timeline has audio and video tracks
(that can handle when we play or not the
audio and video).
- The tracks have parts, that are virtual
items to simplify the way we combine 
tracks. Those parts include media 
instances. The tracks have priorities
ones against the others.
- The media instances have a time range
in which they must be played within the
track (within the timeline). Those media
instances have the start and end time
range for the media source and can apply
effects to the frames.
- The media sources are just the way we
obtain the frames for the specific media
items (read a file, constant color, etc.).
"""
from yta_editor.tracks.items.transition.transition_node_registry import register_transition_processor_nodes


def main():
    from yta_editor.tests import video_modified_stored

    video_modified_stored()


if __name__ == '__main__':
    pass
    # main()

# TODO: This should be somewhere else I think...
# Things I need to execute before launching the system
print('Launching things we need before...')
# TODO: Pass the 'render_context'
# register_transition_processor_nodes()






# TODO: Move this
from yta_editor_common.context import RenderContext
from yta_video_opengl.context.provider import OpenGLContextProvider
from yta_editor.timeline import Timeline
from yta_video_pyav.settings import Settings
from typing import Union
from quicktions import Fraction


class YTAEditor:
    """
    Class to work with our awesome editor.

    TODO: Each 'YTAEditor' should have its own
    `moderngl.Context` I think, so we need to
    change it and make it not singleton.
    """

    def __init__(
        self,
        number_of_tracks: int = 2,
        output_size: tuple[int, int] = Settings.DEFAULT_VIDEO_SIZE.value, # (1920, 1080) before
        fps: Union[int, float, Fraction] = Settings.DEFAULT_VIDEO_FPS.value, # 60.0 before
        # TODO: I don't like this name
        audio_fps: Union[int, Fraction] = Settings.DEFAULT_AUDIO_FPS.value, # 44_100 before # 48_000.0 for aac
        # TODO: Where does this come from (?)
        audio_samples_per_frame: int = Settings.DEFAULT_AUDIO_SAMPLES_PER_FRAME.value, # 1_024 before
        video_codec: str = Settings.DEFAULT_VIDEO_CODEC.value, # h264 before
        video_pixel_format: str = Settings.DEFAULT_PIXEL_FORMAT.value, # yuv420p before
        audio_codec: str = Settings.DEFAULT_AUDIO_CODEC.value, # aac before
        audio_layout: str  = Settings.DEFAULT_AUDIO_LAYOUT.value, # stereo before
        audio_format: str = Settings.DEFAULT_AUDIO_FORMAT.value # fltp before
    ):
        # Validate the parameters and/or set default
        output_size = (
            Settings.DEFAULT_VIDEO_SIZE.value
            if output_size is None else
            output_size
        )

        fps = (
            Settings.DEFAULT_VIDEO_FPS.value
            if fps is None else
            fps
        )

        audio_fps = (
            Settings.DEFAULT_AUDIO_FPS.value
            if audio_fps is None else
            audio_fps
        )

        # TODO: These values below should go in the RenderContext
        audio_samples_per_frame = (
            # 48_000 is not working well, why?
            Settings.DEFAULT_AUDIO_SAMPLES_PER_FRAME.value
            if audio_samples_per_frame is None else
            audio_samples_per_frame
        )

        video_codec = (
            Settings.DEFAULT_VIDEO_CODEC.value
            if video_codec is None else
            video_codec
        )

        video_pixel_format = (
            Settings.DEFAULT_PIXEL_FORMAT.value
            if video_pixel_format is None else
            video_pixel_format
        )

        audio_codec = (
            Settings.DEFAULT_AUDIO_CODEC.value
            if audio_codec is None else
            audio_codec
        )

        audio_layout = (
            Settings.DEFAULT_AUDIO_LAYOUT.value
            if audio_layout is None else
            audio_layout
        )

        audio_format = (
            Settings.DEFAULT_AUDIO_FORMAT.value
            if audio_format is None else
            audio_format
        )

        # Generate a unique context for this YTAEditor instance
        opengl_context = OpenGLContextProvider().context

        # Generate a unique render context for this timeline
        self.render_context = RenderContext(
            opengl_context = opengl_context,
            size = output_size,
            fps = fps,
            audio_fps = audio_fps,
            # TODO: By now I'm forcing this
            do_use_gpu = True
        )

        # TODO: The Timeline should receive the `render_context`
        # instead of these parameters
        self.timeline: Timeline = Timeline(
            render_context = self.render_context,
            number_of_tracks = number_of_tracks,
            # output_size = output_size,
            # fps = fps,
            # audio_fps = audio_fps,
            # TODO: I think all this below should be also in the RenderContext
            audio_samples_per_frame = audio_samples_per_frame,
            video_codec = video_codec,
            video_pixel_format = video_pixel_format,
            audio_codec = audio_codec,
            audio_layout = audio_layout,
            audio_format = audio_format
        )

        # Register transition nodes
        register_transition_processor_nodes(
            render_context = self.render_context
        )



from yta_editor.transformations.effects.video import VideoEffectSingleNodeBaseInput
from yta_editor_nodes.processor.transparency import TransparencyNodeProcessor
from yta_editor_parameters.values.abstract import VideoEditorParameterValue
from yta_editor_parameters.values import ConstantVideoEditorParameterValue
from yta_validation.parameter import ParameterValidator


class TransparencyVideoEffect(VideoEffectSingleNodeBaseInput):
    """
    The effect that will change the transparency of the
    element according to the provided conditions.
    """

    _parameter_values_to_skip_process: dict[str, any] = {
        'transparency': 1.0
    }

    def __init__(
        self,
        render_context: 'RenderContext',
        transparency: VideoEditorParameterValue = ConstantVideoEditorParameterValue(1.0)
    ):
        ParameterValidator.validate_mandatory_instance_of('render_context', render_context, 'RenderContext')
        ParameterValidator.validate_mandatory_instance_of('transparency', transparency, VideoEditorParameterValue)

        super().__init__(
            node_processor = TransparencyNodeProcessor,
            parameters = {
                'transparency': transparency
            },
            render_context = render_context
        )


"""
The items module.

A part is an element that has been placed
in an specific track. We also consider a
part some imaginary things that are 
fulfilling the track regions in which the
user didn't put any media, so we can 
interact with those 'empty' items to 
obtain 'empty' audio and video frames.
"""


from yta_editor.utils import VideoUtils
from yta_editor_frame.frame import Frame
from yta_validation.parameter import ParameterValidator
from av.video.frame import VideoFrame
from av.audio.frame import AudioFrame
from typing import Union

import numpy as np


IS_FROM_GAP_METADATA = 'is_from_gap'
"""
Metadata key to indicate if the frame
has been generated by a gap and should
be ignored when trying to combine with
others.
"""
IS_SILENT_METADATA = 'is_silent'
"""
Metadata key to indicate if the frame is a
silent frame or not, being able to handle it
properly when we need to combine it.
"""

# TODO: These classes will disappear because the
# new 'Frame' and 'AudioFrame' have metadata
class _FrameWrappedBase:
    """
    Class to wrap video and audio frames from
    the pyav library but to support a metadata
    field to inject some information we need
    when processing and combining them.
    """

    @property
    def frame(
        self
    ) -> Frame:
        """
        Get the original pyav VideoFrame wrapped in
        a `Frame` instance.
        """
        return self._frame

    @property
    def is_from_gap(
        self
    ) -> bool:
        """
        Flag to indicate if the frame comes from
        an empty part or not, that will be done
        by checking the 'is_from_gap' 
        attribute in the metadata.

        This will return True only if the value
        is set and it is set as 'True'.
        """
        return self._is_metadata_boolean_set(IS_FROM_GAP_METADATA)
    
    @property
    def is_silent(
        self
    ) -> bool:
        """
        Flag to indicate if the frame that comes is
        actually an empty frame or not, which could
        be useful when we need to combine it with 
        others, and let us avoid reading the content
        to detect if it is silent or not.

        This metadata is valid for audio frames.

        This will return `True` only if the value is
        set and its value is `'True'`.
        """
        return self._is_metadata_boolean_set(IS_SILENT_METADATA)

    def __init__(
        self,
        # TODO: Maybe rename 'Frame' to 'VideoFrameBuffer'
        frame: Union[Frame, AudioFrame],
        metadata: dict = {}
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, [Frame, AudioFrame])
        ParameterValidator.validate_mandatory_dict('metadata', metadata)

        self._frame: Union[Frame, AudioFrame] = frame
        """
        The `Frame` (wrapping a pyav video frame), or
        an `AudioFrame` pyav instance.
        """
        # self.metadata: dict = metadata or {}
        # """
        # The metadata we want to include with the
        # frame.
        # """

    def __repr__(
        self
    ):
        cname = self.__class__.__name__

        return f'<{cname} metadata={self._frame.metadata} frame={self._frame!r}>'
    
    def set_is_from_gap(
        self,
        value: bool = True
    ) -> None:
        """
        Add the metadata information to indicate
        that this is a frame that comes from an 
        empty part.
        """
        return self._set_metadata_boolean(
            attribute_name = IS_FROM_GAP_METADATA,
            value = value
        )

    def set_is_silent(
        self,
        value: bool = True
    ) -> None:
        """
        Add the metadata information to indicate that this
        is a silent frame, including the `is_silent` 
        metadata.
        """
        return self._set_metadata_boolean(
            attribute_name = IS_SILENT_METADATA,
            value = value
        )

    def _is_metadata_boolean_set(
        self,
        attribute_name: str
    ) -> bool:
        """
        *For internal use only*

        Method to simplify the way we check if the boolean
        metadata `attribute_name` is set or not. A metadata
        boolean parameter must be stored as string `'True`'
        to be considered valid.
        """
        # is_parameter = self._frame.metadata.get(attribute_name, None)
        is_parameter = self.metadata.get(attribute_name, None)

        return (
            is_parameter == 'True'
            if is_parameter is not None else
            is_parameter
        )
    
    def _set_metadata_boolean(
        self,
        attribute_name: str,
        value: bool = True
    ) -> None:
        """
        Set the `attribute_name` metadata boolean parameter
        `value` provided (that will be stored as a str).
        """
        value = (
            'True'
            if value else
            'False'
        )

        self._frame.metadata.__setattr__(attribute_name, value)
    
# TODO: Maybe move? But careful with
# cyclic import issues
class AudioFrameWrapped(_FrameWrappedBase):
    """
    Class to wrap audio frames from the pyav
    library but to support a metadata field
    to inject some information we need when
    processing and combining them.
    """

    @property
    def copy(
        self
    ) -> 'AudioFrameWrapped':
        """
        A copy of this same instance.
        """
        return AudioFrameWrapped(
            # TODO: Do I need to hard-copy this '_frame' (?)
            frame = self._frame,
            # metadata = self._frame.metadata
            metadata = self.metadata
        )

    def __init__(
        self,
        frame: AudioFrame,
        metadata: dict = {},
        is_from_gap: bool = False,
        is_silent: bool = False
    ):
        ParameterValidator.validate_mandatory_instance_of('frame', frame, AudioFrame)

        super().__init__(
            frame = frame,
            metadata = metadata
        )

        # TODO: Remoe this when refactored
        self.metadata = {}

        self.set_is_from_gap(is_from_gap)
        self.set_is_silent(is_silent)

    # TODO: This need to be refactored, by now is
    # just to be able to make it work
    def set_is_from_gap(
        self,
        value: bool
    ) -> None:
        self.metadata['is_from_gap'] = (
            'True'
            if value else
            'False'
        )

    def set_is_silent(
        self,
        value: bool
    ) -> None:
        self.metadata['is_silent'] = (
            'True'
            if value else
            'False'
        )


from yta_constants.enum import YTAEnum as Enum


class FrameAdapterMode(Enum):
    """
    The different modes we have to adapt the frames to
    a specific resolution.
    """

    ALPHA_BACKGROUND = 'alpha_background'
    """
    A full transparent frame will be created with the
    expected resolution and the original frame will be
    placed in the middle.
    """